using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Elwig.Helpers;
using Elwig.Models;
using System.Threading.Tasks;
using Microsoft.EntityFrameworkCore.ChangeTracking;

namespace Elwig.Windows {
    public partial class MemberAdminWindow : AdministrationWindow {
        private bool IsEditing = false;
        private bool IsCreating = false;
        private List<string> TextFilter = new();
        private readonly RoutedCommand CtrlF = new();

        public MemberAdminWindow() {
            InitializeComponent();
            CtrlF.InputGestures.Add(new KeyGesture(Key.F, ModifierKeys.Control));
            CommandBindings.Add(new CommandBinding(CtrlF, FocusSearchInput));
            ExemptInputs = new Control[] {
                SearchInput, ActiveMemberInput, MemberList,
                NewMemberButton, EditMemberButton, DeleteMemberButton,
                ResetButton, SaveButton, CancelButton
            };
            RequiredInputs = new Control[] {
                MgNrInput, GivenNameInput, FamilyNameInput,
                AddressInput, PlzInput, OrtInput, BillingOrtInput,
                BusinessSharesInput, BranchInput, DefaultKgInput
            };
        }

        private void Window_Loaded(object sender, RoutedEventArgs evt) {
            ActiveMemberInput.IsChecked = true;
            BranchInput.ItemsSource = Context.Branches.OrderBy(b => b.Name).ToList();
            DefaultKgInput.ItemsSource = Context.WbKgs.Select(k => k.Kg).OrderBy(k => k.Name).ToList();
        }

        private async Task RefreshMemberList() {
            await Context.Members.LoadAsync();
            await RefreshMemberListQuery();
        }

        private async Task RefreshMemberListQuery() {
            IQueryable<Member> memberQuery = Context.Members;
            if (ActiveMemberInput.IsChecked == true)
                memberQuery = memberQuery.Where(m => m.IsActive);

            List<Member> members = await memberQuery.ToListAsync();

            if (TextFilter.Count > 0) {
                members = members
                    .ToDictionary(m => m, m => m.SearchScore(TextFilter))
                    .OrderByDescending(a => a.Value)
                    .ThenBy(a => a.Key.FamilyName)
                    .ThenBy(a => a.Key.GivenName)
                    .Where(a => a.Value > 0)
                    .Select(a => a.Key)
                    .ToList();
            } else {
                members = members
                    .OrderBy(m => m.FamilyName)
                    .ThenBy(m => m.GivenName)
                    .ToList();
            }

            MemberList.ItemsSource = members;
            if (members.Count == 1)
                MemberList.SelectedIndex = 0;

            RefreshInputs();
        }

        private void RefreshInputs(bool validate = false) {
            ClearInputStates();
            Member m = (Member)MemberList.SelectedItem;
            if (m != null) {
                EditMemberButton.IsEnabled = true;
                DeleteMemberButton.IsEnabled = true;
                AreaCommitmentButton.IsEnabled = true;
                FillInputs(m);
            } else {
                EditMemberButton.IsEnabled = false;
                DeleteMemberButton.IsEnabled = false;
                AreaCommitmentButton.IsEnabled = false;
                ClearOriginalValues();
                ClearInputs();
            }
            if (!validate) ClearInputStates();
            GC.Collect();
        }

        private async void InitInputs() {
            MgNrInput.Text = (await Context.NextMgNr()).ToString();
            EntryDateInput.Text = DateTime.Now.ToString("dd.MM.yyyy");
            if (Context.Branches.Count() == 1)
                BranchInput.SelectedItem = Context.Branches.First();
            ActiveInput.IsChecked = true;
            ContactPostInput.IsChecked = true;
            FillOriginalValues();
            ValidateRequiredInputs();
        }

        private void MemberList_SelectionChanged(object sender, RoutedEventArgs evt) {
            RefreshInputs();
        }

        private async void ActiveMemberInput_Changed(object sender, RoutedEventArgs evt) {
            await RefreshMemberListQuery();
        }

        private void NewMemberButton_Click(object sender, RoutedEventArgs evt) {
            IsCreating = true;
            MemberList.IsEnabled = false;
            MemberList.SelectedItem = null;
            HideNewEditDeleteButtons();
            ShowSaveResetCancelButtons();
            UnlockInputs();
            InitInputs();
            LockSearchInputs();
        }

        private void EditMemberButton_Click(object sender, RoutedEventArgs evt) {
            if (MemberList.SelectedItem == null)
                return;

            IsEditing = true;
            MemberList.IsEnabled = false;

            HideNewEditDeleteButtons();
            ShowSaveResetCancelButtons();
            UnlockInputs();
            LockSearchInputs();
        }

        private async void DeleteMemberButton_Click(object sender, RoutedEventArgs evt) {
            Member m = (Member)MemberList.SelectedItem;
            if (m == null) return;

            DeleteMemberButton.IsEnabled = false;
            EditMemberButton.IsEnabled = false;
            var r = MessageBox.Show(
                $"Soll das Mitglied \"{m.FamilyName} {m.GivenName}\" (MgNr. {m.MgNr}) wirklich unwiderruflich gelöscht werden?",
                "Mitglied löschen", MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.No);
            if (r == MessageBoxResult.Yes) {
                Context.Remove(m);
                await Context.SaveChangesAsync();
                await RefreshMemberList();
            }
        }

        private async void SaveButton_Click(object sender, RoutedEventArgs evt) {
            Member m = await UpdateMember(IsEditing ? (Member)MemberList.SelectedItem : Context.CreateProxy<Member>());
            IsEditing = false;
            IsCreating = false;
            MemberList.IsEnabled = true;
            HideSaveResetCancelButtons();
            ShowNewEditDeleteButtons();
            LockInputs();
            UnlockSearchInputs();
            await RefreshMemberList();
            SearchInput.Text = "";
            MemberList.SelectedItem = m;
        }

        private void ResetButton_Click(object sender, RoutedEventArgs evt) {
            if (IsEditing) {
                RefreshInputs();
            } else if (IsCreating) {
                ClearInputs();
                InitInputs();
            }
            UpdateButtons();
        }

        private void CancelButton_Click(object sender, RoutedEventArgs evt) {
            IsEditing = false;
            IsCreating = false;
            MemberList.IsEnabled = true;
            HideSaveResetCancelButtons();
            ShowNewEditDeleteButtons();
            RefreshInputs();
            ClearInputStates();
            LockInputs();
            UnlockSearchInputs();
        }

        private void AreaCommitmentButton_Click(object sender, RoutedEventArgs evt) {
            var w = new AreaComAdminWindow((Member)MemberList.SelectedItem);
            w.Show();
        }

        private async void SearchInput_TextChanged(object sender, RoutedEventArgs evt) {
            TextFilter = SearchInput.Text.ToLower().Split(" ").ToList().FindAll(e => e.Length > 0);
            await RefreshMemberListQuery();
        }

        private void Menu_Member_SendEmail_Click(object sender, RoutedEventArgs evt) {
            Utils.MailTo(((Member)MemberList.SelectedItem).Email);
        }

        private void FocusSearchInput(object sender, RoutedEventArgs evt) {
            if (!IsEditing && !IsCreating) {
                SearchInput.Focus();
                SearchInput.SelectAll();
            }
        }

        private void ShowSaveResetCancelButtons() {
            SaveButton.IsEnabled = false;
            ResetButton.IsEnabled = false;
            CancelButton.IsEnabled = true;
            SaveButton.Visibility = Visibility.Visible;
            ResetButton.Visibility = Visibility.Visible;
            CancelButton.Visibility = Visibility.Visible;
        }

        private void HideSaveResetCancelButtons() {
            SaveButton.IsEnabled = false;
            ResetButton.IsEnabled = false;
            CancelButton.IsEnabled = false;
            SaveButton.Visibility = Visibility.Hidden;
            ResetButton.Visibility = Visibility.Hidden;
            CancelButton.Visibility = Visibility.Hidden;
        }

        private void ShowNewEditDeleteButtons() {
            NewMemberButton.IsEnabled = true;
            EditMemberButton.IsEnabled = MemberList.SelectedItem != null;
            DeleteMemberButton.IsEnabled = MemberList.SelectedItem != null;
            NewMemberButton.Visibility = Visibility.Visible;
            EditMemberButton.Visibility = Visibility.Visible;
            DeleteMemberButton.Visibility = Visibility.Visible;
        }

        private void HideNewEditDeleteButtons() {
            NewMemberButton.IsEnabled = false;
            EditMemberButton.IsEnabled = false;
            DeleteMemberButton.IsEnabled = false;
            NewMemberButton.Visibility = Visibility.Hidden;
            EditMemberButton.Visibility = Visibility.Hidden;
            DeleteMemberButton.Visibility = Visibility.Hidden;
        }

        private void LockSearchInputs() {
            SearchInput.IsEnabled = false;
            ActiveMemberInput.IsEnabled = false;
        }

        private void UnlockSearchInputs() {
            SearchInput.IsEnabled = true;
            ActiveMemberInput.IsEnabled = true;
        }

        private async Task<Member> UpdateMember(Member m) {
            int newMgNr = int.Parse(MgNrInput.Text);
            m.PredecessorMgNr = (PredecessorMgNrInput.Text == "") ? null : int.Parse(PredecessorMgNrInput.Text);
            m.Prefix = (PrefixInput.Text == "") ? null : PrefixInput.Text;
            m.GivenName = GivenNameInput.Text;
            m.FamilyName = FamilyNameInput.Text;
            m.Suffix = (SuffixInput.Text == "") ? null : SuffixInput.Text;
            m.Birthday = (BirthdayInput.Text == "") ? null : string.Join("-", BirthdayInput.Text.Split(".").Reverse());
            m.CountryCode = "AT";
            m.PostalDestId = ((AT_PlzDest)OrtInput.SelectedItem).Id;
            m.Address = AddressInput.Text;

            m.Email = (EmailInput.Text == "") ? null : EmailInput.Text;
            m.PhoneLandline = (PhoneLandlineInput.Text == "") ? null : PhoneLandlineInput.Text.Replace(" ", "");
            m.PhoneMobile1 = (PhoneMobile1Input.Text == "") ? null : PhoneMobile1Input.Text.Replace(" ", "");
            m.PhoneMobile2 = (PhoneMobile2Input.Text == "") ? null : PhoneMobile2Input.Text.Replace(" ", "");

            m.Iban = (IbanInput.Text == "") ? null : IbanInput.Text.Replace(" ", "");
            m.Bic = (BicInput.Text == "") ? null : BicInput.Text;

            m.UstId = (UstIdInput.Text == "") ? null : UstIdInput.Text;
            m.LfbisNr = (LfbisNrInput.Text == "") ? null : LfbisNrInput.Text;
            m.IsBuchführend = BuchführendInput.IsChecked ?? false;

            // TODO Rechnungsadresse

            m.EntryDateString = (EntryDateInput.Text == "") ? null : string.Join("-", EntryDateInput.Text.Split(".").Reverse());
            m.ExitDateString = (ExitDateInput.Text == "") ? null : string.Join("-", ExitDateInput.Text.Split(".").Reverse());
            m.BusinessShares = (BusinessSharesInput.Text == "") ? 0 : int.Parse(BusinessSharesInput.Text);
            m.AccountingNr = (AccountingNrInput.Text == "") ? null : AccountingNrInput.Text;
            m.IsActive = ActiveInput.IsChecked ?? false;
            m.IsVollLieferant = VollLieferantInput.IsChecked ?? false;
            m.IsFunktionär = FunkionärInput.IsChecked ?? false;
            m.ZwstId = ((Branch)BranchInput.SelectedItem).ZwstId;
            m.DefaultKgNr = ((AT_Kg)DefaultKgInput.SelectedItem).KgNr;
            m.Comment = (CommentInput.Text == "") ? null : CommentInput.Text;
            m.DefaultContact = "post";
            if (ContactPostInput.IsChecked ?? false) m.DefaultContact = "post";
            if (ContactEmailInput.IsChecked ?? false) m.DefaultContact = "email";

            EntityEntry<Member>? tr = null;
            try {
                if (IsEditing) {
                    tr = Context.Update(m);
                } else if (IsCreating) {
                    m.MgNr = newMgNr;
                    tr = (await Context.AddAsync(m));
                }
                await Context.SaveChangesAsync();

                if (newMgNr != m.MgNr) {
                    await Context.Database.ExecuteSqlAsync($"UPDATE member SET mgnr = {newMgNr} WHERE mgnr = {m.MgNr}");
                    await Context.Members.LoadAsync();
                    m = await Context.Members.FindAsync(newMgNr);
                }
            } catch (Exception exc) {
                if (tr != null) await tr.ReloadAsync();
                var str = "Der Eintrag konnte nicht in der Datenbank aktualisiert werden!\n\n" + exc.Message;
                if (exc.InnerException != null) str += "\n\n" + exc.InnerException.Message;
                MessageBox.Show(str, "Mitglied aktualisieren", MessageBoxButton.OK, MessageBoxImage.Error);
            }

            return m;
        }

        private void FillInputs(Member m) {
            ClearOriginalValues();

            MgNrInput.Text = m.MgNr.ToString();
            PredecessorMgNrInput.Text = m.PredecessorMgNr.ToString();
            PrefixInput.Text = m.Prefix;
            GivenNameInput.Text = m.GivenName;
            FamilyNameInput.Text = m.FamilyName;
            SuffixInput.Text = m.Suffix;
            BirthdayInput.Text = (m.Birthday != null) ? string.Join(".", m.Birthday.Split("-").Reverse()) : null;
            AddressInput.Text = m.Address;
            AT_PlzDest? p = m.PostalDest.AtPlz;
            if (p != null) {
                PlzInput.Text = p.Plz.ToString();
                OrtInput.ItemsSource = p.AtPlz.Orte;
                OrtInput.SelectedItem = p;
            } else {
                PlzInput.Text = null;
                OrtInput.ItemsSource = null;
                OrtInput.SelectedItem = null;
            }

            EmailInput.Text = m.Email;
            PhoneLandlineInput.Text = m.PhoneLandline;
            PhoneMobile1Input.Text = m.PhoneMobile1;
            PhoneMobile2Input.Text = m.PhoneMobile2;

            IbanInput.Text = m.Iban;
            BicInput.Text = m.Bic;

            UstIdInput.Text = m.UstId;
            LfbisNrInput.Text = m.LfbisNr;
            BuchführendInput.IsChecked = m.IsBuchführend;

            var billingAddr = m.BillingAddress;
            if (billingAddr != null) {
                BillingName.Text = billingAddr.Name;
                BillingAddressInput.Text = billingAddr.Address;
                AT_PlzDest? b = billingAddr.PostalDest.AtPlz;
                if (b != null) {
                    BillingPlzInput.Text = b.Plz.ToString();
                    BillingOrtInput.ItemsSource = b.AtPlz.Orte;
                    BillingOrtInput.SelectedItem = b;
                }
            } else {
                BillingName.Text = "";
                BillingAddressInput.Text = "";
                BillingPlzInput.Text = "";
                BillingOrtInput.ItemsSource = null;
                BillingOrtInput.SelectedItem = null;
            }

            EntryDateInput.Text = (m.EntryDateString != null) ? string.Join(".", m.EntryDateString.Split("-").Reverse()) : null;
            ExitDateInput.Text = (m.ExitDateString != null) ? string.Join(".", m.ExitDateString.Split("-").Reverse()) : null;
            BusinessSharesInput.Text = m.BusinessShares.ToString();
            AccountingNrInput.Text = m.AccountingNr;
            BranchInput.SelectedItem = m.Branch;
            DefaultKgInput.SelectedItem = m.DefaultKg;
            CommentInput.Text = m.Comment;
            ActiveInput.IsChecked = m.IsActive;
            VollLieferantInput.IsChecked = m.IsVollLieferant;
            FunkionärInput.IsChecked = m.IsFunktionär;
            switch (m.DefaultContact) {
                case "post": ContactPostInput.IsChecked = true; break;
                case "email": ContactEmailInput.IsChecked = true; break;
            }

            AreaCommitment.Text = $"{m.ActiveContracts.Select(c => c.Area).Sum():N0} m²";

            Menu_Member_SendEmail.IsEnabled = m.Email != null;

            FillOriginalValues();
        }

        new protected void ClearInputs() {
            Menu_Member_SendEmail.IsEnabled = false;
            AreaCommitment.Text = "- m²";
            base.ClearInputs();
        }

        override protected void UpdateButtons() {
            if (!IsEditing && !IsCreating) return;
            bool ch = HasChanged(), v = IsValid();
            ResetButton.IsEnabled = (ch);
            SaveButton.IsEnabled = (v && ch);
        }

        protected void InputTextChanged(TextBox input, Func<TextBox, bool, AppDbContext, Member?, ValidationResult> checker) {
            InputTextChanged(input, checker(input, SenderIsRequired(input), Context, (Member)MemberList.SelectedItem));
        }

        protected void InputLostFocus(TextBox input, Func<TextBox, bool, AppDbContext, Member?, ValidationResult> checker, string? msg = null) {
            InputLostFocus(input, checker(input, SenderIsRequired(input), Context, (Member)MemberList.SelectedItem), msg);
        }

        private void MgNrInput_TextChanged(object sender, RoutedEventArgs evt) {
            InputTextChanged((TextBox)sender, Validator.CheckMgNr);
        }

        private void MgNrInput_LostFocus(object sender, RoutedEventArgs evt) {
            InputLostFocus((TextBox)sender, Validator.CheckMgNr);
        }

        private void PredecessorMgNrInput_TextChanged(object sender, RoutedEventArgs evt) {
            InputTextChanged((TextBox)sender, Validator.CheckPredecessorMgNr);
        }

        private void PredecessorMgNrInput_LostFocus(object sender, RoutedEventArgs evt) {
            InputLostFocus((TextBox)sender, Validator.CheckPredecessorMgNr);
        }
    }
}
